/* ----------------------------------------------------------------------------
 *  $RCSfile: ipdump.c,v $
 * 
 * $Revision: 1.12 $
 *   $Author: larse $
 *     $Date: 2000/06/07 16:43:18 $
 *    $State: Exp $
 * ----------------------------------------------------------------------------
 * Copyright 1999-2000, Lars Eggert <larse@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the "Artistic License" which comes with this
 * Kit.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License for more details.
 * ----------------------------------------------------------------------------
 *
 * $Log: ipdump.c,v $
 * Revision 1.12  2000/06/07 16:43:18  larse
 * Moved software over to sourceforge.net under the Artistic License. No
 * functional changes to latest release.
 *
 * Revision 1.11  1999/08/13 23:50:10  larse
 * Nicer header separation; fixed quirk with printing newlines.
 *
 * Revision 1.10  1999/08/10 21:34:12  larse
 * Fixed loopback interface dumping.
 *
 * Revision 1.9  1999/05/20 21:20:08  larse
 * Now compiles independent of inet includes. (First step towards
 * multi-platform support.)
 *
 * Revision 1.8  1999/05/18 00:47:24  larse
 * Added ARP/RARP and DVMRP. Factored out common code to basics.c. Added
 * IGMP traceroute (untested).
 *
 * Revision 1.7  1999/03/27 02:44:18  larse
 * Added support for ICMPv4 type 40 (security failure). CAIRN IPsec does
 * not seem to generate those, however. The -v option is now controls
 * more things and has been debugged. Minor cleanup of the code. Added
 * ipd script as a convenient shortcut, it may go away later.
 *
 * Revision 1.6  1999/03/26 00:59:46  larse
 * Now supporting options to inhibit host name lookups, hide link layer
 * headers and show payload data. Documented the code better. Fixed a
 * small bug in IP option handling. First cut at a README.
 *
 * Revision 1.5  1999/03/25 20:23:02  larse
 * Now printing a note when unsupported packet types are
 * encountered. Libpcap location now configurable, this allows
 * compilation with CAIRN's broken libpcap (fixed one in
 * /home/d7unsupp/FreeBSD). Minor other changes.
 *
 * Revision 1.4  1999/03/24 00:29:56  larse
 * Added support for IPIP tunnels, IPsec (AH and ESP) and raw data
 * output. Fixed bugs in the ICMPv4 module. Fixed bug printing certain
 * options in the IP module. Minor changes to other modules, mostly to
 * write raw data where it was omitted previously.
 *
 * Revision 1.3  1999/03/23 03:07:06  larse
 * Added support for TCP and UDP. Cleaned up IP option handling and ICMP.
 *
 * Revision 1.2  1999/03/22 20:20:21  larse
 * First stable and partially working version. IPv4 and ICMPv4 are the
 * only supported protocols, but they work fine, even with
 * extensions. The reorganized code should allow for much easier
 * extensions for other protocols.
 *
 * Revision 1.1  1999/03/19 04:22:48  larse
 * Initial revision
 *
 * ----------------------------------------------------------------------------
 */


/* Define a marker for ident(1). */
static const char rcsid[] = 
  "$Id: ipdump.c,v 1.12 2000/06/07 16:43:18 larse Exp $";


#include <stdio.h>
#include <pcap.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>

#include "ipdump.h"
#include "scan.h"
#include "indent.h"


/* documented in ipdump.h */
const char *exec_name;
u_char unsupported = 0;
#if 0
u_char dump_unsupported = 0;
#endif
u_char verbose = 0;
u_char no_names = 0;
u_char print_ll_hdr = 0;


/* Called upon exit (either normal or by signal.) */
static void cleanup(int signal)
  {
  if(unsupported)
    /* we encountered unsupported header elements, tell the user that he can
       help improve ipdump */
    fprintf(stderr, 
      	    "\n\nSOME PACKETS ANALYZED IN THIS RUN WERE CONTAINING "
      	    "UNSUPPORTED ELEMENTS.\n\nTo help improve ipdump, please consider "
      	    "sending a savefile containing\nthese packets to larse@isi.edu "
      	    "together with a brief explanation on how you \ncaptured them. "
      	    "Please include hints (URLs, etc.) describing the packet\nformat, "
      	    "if available. You may also send me a patch, of course...\n\n");
  exit(0);
  }


/* print an IP packet */
static void print_ip(void)
  {
  const u_char *type = next(sizeof(u_char));
  if(type == 0) return;

  /* see if we deal with an IPv4 or IPv6 packet */
  switch(*type >> 4)
    {
    case 4: unget(sizeof(u_char)); print_ipv4(1); break;
    default:
      unsupported++;
      fprintf(stderr, "%s: packet has unsupported IP version %hu, ignoring\n", 
	      exec_name, *type >> 4);
    }
  }


/* pcap packet handler, called for each packet by libpcap */
static void print_pkt(u_char *u, const struct pcap_pkthdr *h, const u_char *p)
  {
#if 0
  const u_long old_unsupported = unsupported;
#endif

  /* start dump with no indentation, init the header scanner */
  indent_level = 0;
  init_scanner(h->caplen, p);
  
  /* print data link layer header */
  switch(pcap_datalink((pcap_t *)u))
    {
    case DLT_NULL:
      /* no link-layer encapsulation, eat a long, don't know why */
      {
      const u_long *dummy = (u_long *)next(sizeof(u_long));
      if(dummy == 0) return;
      print_ip();
      break;
      }
      
    case DLT_EN10MB:
      /* Ethernet (10Mb) */
    case DLT_EN3MB:
      /* Experimental Ethernet (3Mb) */
      print_ether();
      break;
      
    case DLT_AX25:
      /* Amateur Radio AX.25 */
    case DLT_PRONET:
      /* Proteon ProNET Token Ring */
    case DLT_CHAOS:
      /* Chaos */
    case DLT_IEEE802:
      /* IEEE 802 Networks */
    case DLT_ARCNET:
      /* ARCNET */
    case DLT_SLIP:
      /* Serial Line IP */
    case DLT_PPP:
      /* Point-to-point Protocol */
    case DLT_FDDI:
      /* FDDI */
    case DLT_ATM_RFC1483:
      /* LLC/SNAP encapsulated atm */
    default:
      unsupported++;
      fprintf(stderr, "%s: unsupported link layer type %hu\n", exec_name, 
	      pcap_datalink((pcap_t *)u));
    }
#if 0
  /* if this packet has unsupported elements, dump it */
  if(dump_unsupported && old_unsupported < unsupported) 
    {
    pcap_dumper_t *pd = pcap_dump_open((pcap_t *)u, "a");
    pcap_dump(u, h, p);
    pcap_dump_close(pd);
    }
#endif
  }


static void usage(void)
  {
  fprintf(stderr, "usage: %s [-env] < file\n", exec_name);
  fprintf(stderr, "\t-e  print link-layer header\n");
  fprintf(stderr, "\t-n  suppress all name lookups\n");
  fprintf(stderr, "\t-v  verbose output (print payload data)\n");
#if 0
  fprintf(stderr, "\t-u  dump packets with unsupported fields to stdout\n");
#endif
  cleanup(0);
  }


int main(int argc, char * argv[])
  {
  char ebuf[PCAP_ERRBUF_SIZE];
  pcap_t *pcap;
  int opt;

  /* remember our program name */
  exec_name = argv[0];

  /* we need to do some stuff when exiting, register signal handler */
  signal(SIGHUP, cleanup);
  signal(SIGINT, cleanup);
  signal(SIGKILL, cleanup);
  signal(SIGTERM, cleanup);

#if 0
  while((opt = getopt(argc, argv, "envuh")) != -1)
#endif
  while((opt = getopt(argc, argv, "envh")) != -1)
    switch(opt) 
      {
      case 'e':	print_ll_hdr = 1; break;
      case 'n':	no_names = 1; break;
      case 'v':	verbose = 1; break;
#if 0
      case 'u': dump_unsupported = 1; break;
#endif
      case '?':
      case 'h':
      default:
	usage();
      }
  argc -= optind;
  argv += optind;
  if(argc > 0) usage();

  /* have pcap read from stdin */
  pcap = pcap_open_offline("-", ebuf);

  /* process packets */
  if(pcap) 
    {
    fprintf(stderr, "%s: input produced by libpcap %hu.%hu", exec_name,
	    pcap_major_version(pcap), pcap_minor_version(pcap));
    if(pcap_is_swapped(pcap)) fprintf(stderr, " (different byte-order)");
    fprintf(stderr, "\n");
    if(pcap_loop(pcap, -1, print_pkt, (u_char *)pcap) < 0) 
      pcap_perror(pcap, (char *)exec_name);
    cleanup(0);
    }
  else fprintf(stderr, "%s: %s\n", exec_name, ebuf);
  return 0;
  }
