/* ----------------------------------------------------------------------------
 *  $RCSfile: ipv4.c,v $
 * 
 * $Revision: 1.13 $
 *   $Author: larse $
 *     $Date: 2000/06/07 16:43:18 $
 *    $State: Exp $
 * ----------------------------------------------------------------------------
 * Copyright 1999-2000, Lars Eggert <larse@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the "Artistic License" which comes with this
 * Kit.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License for more details.
 * ----------------------------------------------------------------------------
 *
 * $Log: ipv4.c,v $
 * Revision 1.13  2000/06/07 16:43:18  larse
 * Moved software over to sourceforge.net under the Artistic License. No
 * functional changes to latest release.
 *
 * Revision 1.12  1999/08/13 23:50:10  larse
 * Nicer header separation; fixed quirk with printing newlines.
 *
 * Revision 1.11  1999/08/10 21:35:28  larse
 * Fixes for little-endian machines.
 *
 * Revision 1.10  1999/05/20 21:20:09  larse
 * Now compiles independent of inet includes. (First step towards
 * multi-platform support.)
 *
 * Revision 1.9  1999/05/18 00:47:25  larse
 * Added ARP/RARP and DVMRP. Factored out common code to basics.c. Added
 * IGMP traceroute (untested).
 *
 * Revision 1.8  1999/05/12 13:05:00  larse
 * Added support for SAP and SAP. Removed duplicate includes from other
 * sources. Changed output of IP router alert option in last version, but
 * forgot to document.
 *
 * Revision 1.7  1999/05/11 17:45:06  larse
 * Initial support for IGMPv4, mostly missing is support for DVMRP.
 *
 * Revision 1.6  1999/03/27 02:44:19  larse
 * Added support for ICMPv4 type 40 (security failure). CAIRN IPsec does
 * not seem to generate those, however. The -v option is now controls
 * more things and has been debugged. Minor cleanup of the code. Added
 * ipd script as a convenient shortcut, it may go away later.
 *
 * Revision 1.5  1999/03/26 00:59:47  larse
 * Now supporting options to inhibit host name lookups, hide link layer
 * headers and show payload data. Documented the code better. Fixed a
 * small bug in IP option handling. First cut at a README.
 *
 * Revision 1.4  1999/03/25 20:23:03  larse
 * Now printing a note when unsupported packet types are
 * encountered. Libpcap location now configurable, this allows
 * compilation with CAIRN's broken libpcap (fixed one in
 * /home/d7unsupp/FreeBSD). Minor other changes.
 *
 * Revision 1.3  1999/03/24 00:29:57  larse
 * Added support for IPIP tunnels, IPsec (AH and ESP) and raw data
 * output. Fixed bugs in the ICMPv4 module. Fixed bug printing certain
 * options in the IP module. Minor changes to other modules, mostly to
 * write raw data where it was omitted previously.
 *
 * Revision 1.2  1999/03/23 03:07:07  larse
 * Added support for TCP and UDP. Cleaned up IP option handling and ICMP.
 *
 * Revision 1.1  1999/03/22 20:20:22  larse
 * First stable and partially working version. IPv4 and ICMPv4 are the
 * only supported protocols, but they work fine, even with
 * extensions. The reorganized code should allow for much easier
 * extensions for other protocols.
 *
 * ----------------------------------------------------------------------------
 */


/* Define a marker for ident(1). */
static const char rcsid[] = 
  "$Id: ipv4.c,v 1.13 2000/06/07 16:43:18 larse Exp $";


#include <stdio.h>

#include "ipdump.h"
#include "indent.h"
#include "scan.h"


/* local header id */
static const char hdr[] = "IPv4";


struct ip 
  {
  u_int   ip_hl:4,                /* version */
          ip_v:4;                 /* header length */
  u_char  ip_tos;                 /* type of service */
  u_short ip_len;                 /* total length */
  u_short ip_id;                  /* identification */
  u_short ip_off;                 /* fragment offset field */
  u_char  ip_ttl;                 /* time to live */
  u_char  ip_p;                   /* protocol */
  u_short ip_sum;                 /* checksum */
  u_long  ip_src, ip_dst;	  /* source and dest address */
  };


#define IP_RF 0x8000                    /* reserved fragment flag */
#define IP_DF 0x4000                    /* dont fragment flag */
#define IP_MF 0x2000                    /* more fragments flag */
#define IP_OFFMASK 0x1fff               /* mask for fragmenting bits */

#define IPOPT_COPIED(o)         ((o)&0x80)
#define IPOPT_CLASS(o)          ((o)&0x60)
#define IPOPT_NUMBER(o)         ((o)&0x1f)

#define IPOPT_CONTROL           0x00
#define IPOPT_RESERVED1         0x20
#define IPOPT_DEBMEAS           0x40
#define IPOPT_RESERVED2         0x60

#define IPOPT_EOL               0               /* end of option list */
#define IPOPT_NOP               1               /* no operation */

#define IPOPT_RR                7               /* record packet route */
#define IPOPT_TS                68              /* timestamp */
#define IPOPT_SECURITY          130             /* provide s,c,h,tcc */
#define IPOPT_LSRR              131             /* loose source route */
#define IPOPT_SATID             136             /* satnet id */
#define IPOPT_SSRR              137             /* strict source route */
#define IPOPT_RA                148             /* router alert */

#define IPTOS_LOWDELAY          0x10
#define IPTOS_THROUGHPUT        0x08
#define IPTOS_RELIABILITY       0x04
#define IPTOS_MINCOST           0x02

#define IPTOS_PREC_NETCONTROL           0xe0
#define IPTOS_PREC_INTERNETCONTROL      0xc0
#define IPTOS_PREC_CRITIC_ECP           0xa0
#define IPTOS_PREC_FLASHOVERRIDE        0x80
#define IPTOS_PREC_FLASH                0x60
#define IPTOS_PREC_IMMEDIATE            0x40
#define IPTOS_PREC_PRIORITY             0x20
#define IPTOS_PREC_ROUTINE              0x00

#define IPPROTO_ICMP            1               /* control message protocol */
#define IPPROTO_IGMP            2               /* group mgmt protocol */
#define IPPROTO_IPIP            4               /* IP encapsulation in IP */
#define IPPROTO_TCP             6               /* tcp */
#define IPPROTO_UDP             17              /* user datagram protocol */
#define IPPROTO_ESP             50              /* SIPP Encap Sec. Payload */
#define IPPROTO_AH              51              /* SIPP Auth Header */


/* print route option */
static void print_ipv4_route(void)
  {
  u_short opt_len;
  u_short total_opt_len;
  const u_char *rr_ptr;
  const u_char *oct = next(sizeof(u_char));

  if(oct == 0) return;  
  iprintf(hdr, "length: %hu [octet%s]\n", *oct, *oct == 1 ? "" : "s");
  total_opt_len = opt_len = *oct;
  opt_len -= 2 * sizeof(u_char);

  /* get ptr to last hop */
  rr_ptr = next(sizeof(u_char));
  if(rr_ptr == 0) return;
  opt_len -= sizeof(u_char);
  
  iprintf(hdr, "pointer: %hu\n", *rr_ptr);
  iprintf(hdr, "route:\n");
  indent_level++;

  /* print each entry in the route area */
  while(opt_len > 0)
    {
    const u_long *rr_addr = (const u_long *)next(sizeof(u_long));
    if(rr_addr == 0) return;
    opt_len -= sizeof(u_long);
    print_address(hdr, 0, rr_addr);
    printf("\n");
    }
  indent_level--;
  /* dump the unused route area for the curious */
  if(opt_len > 0)
    {
    iprintf(hdr, "unused route data area: [%hu octet%s]", opt_len, 
	    opt_len == 1 ? "" : "s");
    if(verbose)
      {
      /* print unused route area contents */
      indent_level++;
      print_raw(hdr, opt_len);
      indent_level--;
      }
    else 
      {
      /* skip unused route area */
      const u_char *skip = next(opt_len);
      printf(" [omitted]\n");
      skip = 0;
      }
    }
  }


/* print a record route ip option, return option length */
static void print_ipv4_rr_route(void)
  {
  u_short rr;
  u_short opt_len;
  u_short total_opt_len;
  const u_char *rr_ptr;
  const u_char *oct = next(sizeof(u_char));

  if(oct == 0) return;  
  iprintf(hdr, "length: %hu [octet%s]\n", *oct, *oct == 1 ? "" : "s");
  total_opt_len = opt_len = *oct;
  opt_len -= 2 * sizeof(u_char);

  /* get ptr to first free slot */
  rr_ptr = next(sizeof(u_char));
  if(rr_ptr == 0) return;
  opt_len -= sizeof(u_char);
  
  iprintf(hdr, "pointer: %hu\n", *rr_ptr);
  iprintf(hdr, "route:\n");
  indent_level++;
  
  /* print each entry in the route area before the ptr */
  for(rr = 0; rr < *rr_ptr - 4; rr += sizeof(u_long))
    {
    const u_long *rr_addr = (const u_long *)next(sizeof(u_long));
    if(rr_addr == 0) return;
    opt_len -= sizeof(u_long);
    print_address(hdr, 0, rr_addr);
    printf("\n");
    }
  indent_level--;
  /* dump the unused route area for the curious */
  iprintf(hdr, "unused route data area: [%hu octet%s]", opt_len, 
	  opt_len == 1 ? "" : "s");
  if(verbose) 
    {
    /* print unused route area contents */
    indent_level++;
    print_raw(hdr, opt_len);
    indent_level--;
    }
  else 
    {
    /* skip unused route area */
    const u_char *skip = next(opt_len);
    printf(" [omitted]\n");
    skip = 0;
    }
  }


/* print a generic ip option, return option length */
static void print_ipv4_generic_option(void)
  {
  int opt_len;
  const u_char *oct = next(sizeof(u_char));

  if(oct == 0) return;
  opt_len = *oct;
  iprintf(hdr, "length: %hu [octet%s]\n", *oct, *oct == 1 ? "" : "s");
  iprintf(hdr, "data: [%hu octet%s]", opt_len - 2, 
	  opt_len - 2 == 1 ? "" : "s");
  if(verbose) 
    {
    /* print option data */
    indent_level++;
    print_raw(hdr, opt_len - 2);
    indent_level--;
    }
  else 
    {
    /* skip option data */
    const u_char *skip = next(opt_len - 2);
    printf(" [omitted]\n");
    skip = 0;
    }
  }


/* print a generic ip option, return option length */
static void print_ipv4_router_alert(void)
  {
  const u_char *oct = next(sizeof(u_char));

  if(oct == 0) return;
  iprintf(hdr, "length: %hu [octet%s]\n", *oct, *oct == 1 ? "" : "s");
  oct = next(sizeof(u_char));
  iprintf(hdr, "value: %hu ", ntohs(*oct));
  if(*oct == 0) printf("[router alert]\n");
  else
    {
    unsupported++;
    printf("[*unsupported*]\n");
    }
  }


/* print ip options */
static void print_ipv4_options(const u_short remaining_hdr)
  {
  const u_char *type;
  const u_long pos = position();

  do
    {
    /* get first byte of option (contains type, etc.) */
    type = next(sizeof(u_char));
    if(type == 0) return;

    iprintf(hdr, "option: %hu\n", *type);
    indent_level++;
    iprintf(hdr, "flag: %hu [%scopied]\n", IPOPT_COPIED(*type) ? 1 : 0, 
	    IPOPT_COPIED(*type) ? "" : "not ");
    iprintf(hdr, "class: %hu ", IPOPT_CLASS(*type));
    switch(IPOPT_CLASS(*type))
      {
      case IPOPT_CONTROL: printf("[control]\n"); break;
      case IPOPT_DEBMEAS: printf("[debugging and measurement]\n"); break;
      case IPOPT_RESERVED1: 
      case IPOPT_RESERVED2: printf("[reserved]\n"); break;
      default: 
	unsupported++;
	printf("[*unsupported*]\n");
      }
    iprintf(hdr, "number: %hu ", IPOPT_NUMBER(*type));

    /* based on the type, print remainder */
    switch(*type)
      {
      case IPOPT_EOL:
	printf("[end of option list]\n");
	break;

      case IPOPT_NOP:
	printf("[no operation]\n");
	break;
	
      case IPOPT_RR:
	printf("[record packet route]\n");
	print_ipv4_rr_route();
	break;
	
      case IPOPT_LSRR:
	printf("[loose source and record route]\n");
	print_ipv4_route();
	break;

      case IPOPT_SSRR:
	printf("[strict source and record route]\n");
	print_ipv4_route();
	break;

      case IPOPT_TS:
	printf("[timestamp]\n");
	print_ipv4_generic_option();
	break;
	
      case IPOPT_SECURITY:
	printf("[security]\n");
	print_ipv4_generic_option();
	break;
	
      case IPOPT_SATID:
	printf("[satnet id]\t");
	print_ipv4_generic_option();
	break;
	
      case IPOPT_RA:
	printf("[router alert]\n");
	print_ipv4_router_alert();
	break;
	
      default:
	unsupported++;
	printf("[*unsupported*]\n");
	print_ipv4_generic_option();
      }
    indent_level--;
    } while((position() - pos) < remaining_hdr && 
	    IPOPT_NUMBER(*type) != IPOPT_EOL);
  }


/* print nect header based on protocol */
void print_ipv4_content(const u_char protocol, const u_long destination)
  {
  print_seperator();
  switch(protocol)
    {
    /* see /usr/include/netinet/in.h and RFC 1700 */
    case IPPROTO_ICMP: print_icmpv4(); break;
    case IPPROTO_IGMP: print_igmpv4(); break;  
    case IPPROTO_UDP: print_udp(destination); break;
    case IPPROTO_TCP: print_tcp(); break;
    case IPPROTO_ESP: print_esp(); break;
    case IPPROTO_AH: print_ah(destination); break;
    case IPPROTO_IPIP: print_ipv4(1); break;
    case 103: /*PIM*/
    default:
      unsupported++;
      fprintf(stderr, "%s: IP protocol %hu yet unsupported\n", exec_name, 
	      protocol);
    }
  }


/* print an IPv4 header */
void print_ipv4(const short recurse)
  {
  int padding = 0;
  int fl = 0;
  const u_char TOS_FLAGS = 
    IPTOS_LOWDELAY|IPTOS_THROUGHPUT|IPTOS_RELIABILITY|IPTOS_MINCOST;
  const u_long ip_start = position();
  const struct ip *ip = (const struct ip *)next(sizeof(struct ip));

  if(ip == 0) return;
  iprintf(hdr, "version: %hu\n", ip->ip_v);
  iprintf(hdr, "header length: %hu [%hu octets]\n", ip->ip_hl, ip->ip_hl * 4);
  iprintf(hdr, "type of service: %hu\n", ip->ip_tos);
  indent_level++;
  iprintf(hdr, "precedence: %hu ", ip->ip_tos & 0x7);
  switch(ip->ip_tos & 0x7)
    {
    case IPTOS_PREC_ROUTINE: printf("[routine]\n"); break;
    case IPTOS_PREC_PRIORITY: printf("[priority]\n"); break;
    case IPTOS_PREC_IMMEDIATE: printf("[immediate]\n"); break;
    case IPTOS_PREC_FLASH: printf("[flash]\n"); break;
    case IPTOS_PREC_FLASHOVERRIDE: printf("[flash override]\n"); break;
    case IPTOS_PREC_CRITIC_ECP: printf("[critic/ecp]\n"); break;
    case IPTOS_PREC_INTERNETCONTROL: printf("[internetwork control]\n"); break;
    case IPTOS_PREC_NETCONTROL: printf("[network control]\n"); break;
    default:
      unsupported++;
      printf("[*unsupported*]\n");
    }
  
  iprintf(hdr, "flags: %hu ", ip->ip_tos & TOS_FLAGS);
  if(ip->ip_tos & TOS_FLAGS)
    {
    printf(" [");
    if(ip->ip_tos & IPTOS_LOWDELAY) printf("%slow delay", fl++ ? "|" : "");
    if(ip->ip_tos & IPTOS_THROUGHPUT) 
      printf("%shigh throughput", fl++ ? "|" : "");
    if(ip->ip_tos & IPTOS_RELIABILITY) 
      printf("%shigh reliability", fl++ ? "|" : "");
    /* these clash with IPTOS_MINCOST:
    if(ip->ip_tos & IPTOS_ECT) 
      printf("%secn capable transport", fl++ ? "|" : "");
    if(ip->ip_tos & IPTOS_CE) 
      printf("%scongestion experienced", fl++ ? "|" : "");
    */
    if(ip->ip_tos & IPTOS_MINCOST) printf("%sminimal cost", fl++ ? "|" : "");
    printf("]\n");
    }
  else printf(" [normal]\n");
  indent_level--;
  iprintf(hdr, "total length: %hu [octets]\n", ntohs(ip->ip_len));
  iprintf(hdr, "identification: %hu\n", ntohs(ip->ip_id));
  fl = 0;
  iprintf(hdr, "flags: %hu [", ip->ip_off & (IP_RF|IP_DF|IP_MF));
  if(ip->ip_off & IP_RF) printf("%sreserved", fl++ ? "|" : "");
  if(ip->ip_off & IP_DF) printf("%sdon't fragment", fl++ ? "|" : "");
  else  printf("%smay fragment", fl++ ? "|" : "");
  if(ip->ip_off & IP_MF) printf("%smore fragments", fl++ ? "|" : "");
  else printf("%slast fragment", fl++ ? "|" : "");
  printf("]\n");
  iprintf(hdr, "fragment offset: %hu\n", ntohs(ip->ip_off) & IP_OFFMASK);
  iprintf(hdr, "time to live: %hu\n", ip->ip_ttl);
  print_protocol(hdr, "protocol", ip->ip_p);
  printf("\n");
  iprintf(hdr, "checksum: %hu\n", ntohs(ip->ip_sum));
  print_address(hdr, "source address", (u_long *)&ip->ip_src);
  printf("\n");
  print_address(hdr, "destination address", (u_long *)&ip->ip_dst);
  printf("\n");

  /* print options, if any */
  if(ip->ip_hl * 4 > 20) print_ipv4_options(ip->ip_hl * 4 - 20);

  /* print padding */
  padding = (position() - ip_start) % 4;
  if(padding)
    {
    iprintf(hdr, "padding: [%hu octet%s]", padding, padding == 1 ? "" : "s");
    if(verbose)
      {
      /* print padding contents */
      indent_level++;
      print_raw(hdr, padding);
      indent_level--;
      }
    else 
      {
      /* skip padding */
      const u_char *skip = next(padding);
      printf(" [omitted]\n");
      skip = 0;
      }
    }

  /* continue parsing inner headers (based on protocol type) */
  if(recurse) print_ipv4_content(ip->ip_p, ip->ip_dst);
  }


