/* ----------------------------------------------------------------------------
 *  $RCSfile: tcp.c,v $
 * 
 * $Revision: 1.9 $
 *   $Author: larse $
 *     $Date: 2000/06/07 16:43:19 $
 *    $State: Exp $
 * ----------------------------------------------------------------------------
 * Copyright 1999-2000, Lars Eggert <larse@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the "Artistic License" which comes with this
 * Kit.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License for more details.
 * ----------------------------------------------------------------------------
 *
 * $Log: tcp.c,v $
 * Revision 1.9  2000/06/07 16:43:19  larse
 * Moved software over to sourceforge.net under the Artistic License. No
 * functional changes to latest release.
 *
 * Revision 1.8  1999/08/10 21:35:28  larse
 * Fixes for little-endian machines.
 *
 * Revision 1.7  1999/05/20 21:20:09  larse
 * Now compiles independent of inet includes. (First step towards
 * multi-platform support.)
 *
 * Revision 1.6  1999/05/18 00:47:25  larse
 * Added ARP/RARP and DVMRP. Factored out common code to basics.c. Added
 * IGMP traceroute (untested).
 *
 * Revision 1.5  1999/03/27 02:44:19  larse
 * Added support for ICMPv4 type 40 (security failure). CAIRN IPsec does
 * not seem to generate those, however. The -v option is now controls
 * more things and has been debugged. Minor cleanup of the code. Added
 * ipd script as a convenient shortcut, it may go away later.
 *
 * Revision 1.4  1999/03/26 00:59:48  larse
 * Now supporting options to inhibit host name lookups, hide link layer
 * headers and show payload data. Documented the code better. Fixed a
 * small bug in IP option handling. First cut at a README.
 *
 * Revision 1.3  1999/03/25 20:23:03  larse
 * Now printing a note when unsupported packet types are
 * encountered. Libpcap location now configurable, this allows
 * compilation with CAIRN's broken libpcap (fixed one in
 * /home/d7unsupp/FreeBSD). Minor other changes.
 *
 * Revision 1.2  1999/03/24 00:29:59  larse
 * Added support for IPIP tunnels, IPsec (AH and ESP) and raw data
 * output. Fixed bugs in the ICMPv4 module. Fixed bug printing certain
 * options in the IP module. Minor changes to other modules, mostly to
 * write raw data where it was omitted previously.
 *
 * Revision 1.1  1999/03/23 03:07:07  larse
 * Added support for TCP and UDP. Cleaned up IP option handling and ICMP.
 *
 * ----------------------------------------------------------------------------
 */


/* Define a marker for ident(1). */
static const char rcsid[] = 
  "$Id: tcp.c,v 1.9 2000/06/07 16:43:19 larse Exp $";


#include <stdio.h>

#include "ipdump.h"
#include "indent.h"
#include "scan.h"

struct tcphdr {
        u_short th_sport;               /* source port */
        u_short th_dport;               /* destination port */
        u_long th_seq;                  /* sequence number */
        u_long th_ack;                  /* acknowledgement number */
        u_int   th_x2:4,                /* data offset */
                th_off:4;               /* (unused) */
        u_char  th_flags;
#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PUSH 0x08
#define TH_ACK  0x10
#define TH_URG  0x20
#if 1 /* ALTQ_ECN */
#define TH_ECNECHO      0x40            /* ecn echo */
#define TH_CWR          0x80            /* congestion window reduced */
#define TH_FLAGS (TH_FIN|TH_SYN|TH_RST|TH_ACK|TH_URG|TH_ECNECHO|TH_CWR)
#else
#define TH_FLAGS (TH_FIN|TH_SYN|TH_RST|TH_ACK|TH_URG)
#endif

        u_short th_win;                 /* window */
        u_short th_sum;                 /* checksum */
        u_short th_urp;                 /* urgent pointer */
};

#define TCPOPT_EOL              0
#define TCPOPT_NOP              1
#define TCPOPT_MAXSEG           2
#define    TCPOLEN_MAXSEG               4
#define TCPOPT_WINDOW           3
#define    TCPOLEN_WINDOW               3
#define TCPOPT_SACK_PERMITTED   4               /* Experimental */
#define    TCPOLEN_SACK_PERMITTED       2
#define TCPOPT_SACK             5               /* Experimental */
#define TCPOPT_TIMESTAMP        8
#define    TCPOLEN_TIMESTAMP            10
#define    TCPOLEN_TSTAMP_APPA          (TCPOLEN_TIMESTAMP+2) /* appendix A */
#define    TCPOPT_TSTAMP_HDR            \
    (TCPOPT_NOP<<24|TCPOPT_NOP<<16|TCPOPT_TIMESTAMP<<8|TCPOLEN_TIMESTAMP)

#define TCPOPT_CC               11              /* CC options: RFC-1644 */
#define TCPOPT_CCNEW            12
#define TCPOPT_CCECHO           13
#define    TCPOLEN_CC                   6
#define    TCPOLEN_CC_APPA              (TCPOLEN_CC+2)
#define    TCPOPT_CC_HDR(ccopt)         \
    (TCPOPT_NOP<<24|TCPOPT_NOP<<16|(ccopt)<<8|TCPOLEN_CC)


static const char hdr[] = "TCP";


/* print a generic ttcp option (they all have the same form) */
static void print_tcp_ttcp_option(void)
  {
  const u_long *cc;
  const u_char *oct = next(sizeof(u_char));

  if(oct == 0) return;
  iprintf(hdr, "length: %u [octet%s]\n", *oct, *oct == 1 ? "" : "s");
  cc = (const u_long *)next(sizeof(u_long));
  if(cc == 0) return;  
  iprintf(hdr, "connection count: %u\n", ntohl(*cc));
  }
  

/* print a generic tcp option */
static void print_tcp_generic_tcp_option(void)
  {
  const u_char *oct;
  int opt_len;

  oct = next(sizeof(u_char));
  if(oct == 0) return;
  opt_len = *oct - 2;
  iprintf(hdr, "length: %u [octet%s]\n", *oct, *oct == 1 ? "" : "s");

  iprintf(hdr, "data: [%hu octet%s]", opt_len, opt_len == 1 ? "" : "s");
  if(verbose)
    {
    /* print option data */
    indent_level++;
    print_raw(hdr, opt_len);
    indent_level--;
    }
  else 
    {
    /* skip option data */
    const u_char *skip = next(opt_len);
    printf(" [omitted]\n");
    skip = 0;
    }
  }


/* print all tcp options */
static void print_tcp_options(const short remaining_hdr)
  {
  const u_char *type;
  const u_long pos = position();

  do
    {
    type = next(sizeof(u_char));
    if(type == 0) return;

    /* get option type and print based on it */
    iprintf(hdr, "option: %u ", *type);
    indent_level++;
    switch(*type)
      {
      case TCPOPT_NOP: 
	printf("[no operation]\n"); 
	break;

      case TCPOPT_MAXSEG:
	{
	const u_short *mss;
	const u_char *oct = next(sizeof(u_char));
	if(oct == 0) return;
	printf("[maximum segment size]\n");
	iprintf(hdr, "length: %u [octet%s]\n", *oct, *oct == 1 ? "" : "s");
	mss = (u_short *)next(sizeof(u_short));
	if(mss == 0) return;
	iprintf(hdr, "maximum segment size: %u [octet%s]\n", ntohs(*mss),
		ntohs(*mss) == 1 ? "" : "s");
	break;
	}

      case TCPOPT_WINDOW:
	{
	const u_char *oct = next(sizeof(u_char));
	if(oct == 0) return;
	printf("[window scale factor]\n");
	iprintf(hdr, "length: %u [octet%s]\n", *oct, *oct == 1 ? "" : "s");
	oct = next(sizeof(u_char));
	if(oct == 0) return;
	iprintf(hdr, "shift count: %u\n", *oct);
	break;
	}

      case TCPOPT_TIMESTAMP:
	{
	const u_long *lng;
	const u_char *oct = next(sizeof(u_char));
	if(oct == 0) return;
	printf("[timestamp]\n");
	iprintf(hdr, "length: %u [octet%s]\n", *oct, *oct == 1 ? "" : "s");
	lng = (u_long *)next(sizeof(u_long));
	if(lng == 0) return;
	iprintf(hdr, "timestamp: %lu\n", *lng);
	lng = (u_long *)next(sizeof(u_long));
	if(lng == 0) return;
	iprintf(hdr, "timestamp echo reply: %lu\n", *lng);
	break;
	}

      case TCPOPT_SACK_PERMITTED:
	{
	const u_char *oct = next(sizeof(u_char));
	if(oct == 0) return;
	printf("[sack permitted]\n");
	iprintf(hdr, "length: %u [octet%s]\n", *oct, *oct == 1 ? "" : "s");
	break;
	}

      case TCPOPT_SACK:
	{
	u_short opt_len;
	const u_char *oct = next(sizeof(u_char));
	if(oct == 0) return;
	printf("[sack]\n");
	opt_len = *oct;
	iprintf(hdr, "length: %u [octet%s]\n", *oct, *oct == 1 ? "" : "s");
	iprintf(hdr, "sack blocks:\n");
	indent_level++;
	for(opt_len = *oct; opt_len > 0; opt_len -= 2*sizeof(u_long))
	  {
	  const u_long *lng = (u_long *)next(sizeof(u_long));
	  if(lng == 0) return;
	  iprintf(hdr, "%lu", *lng);
	  lng = (u_long *)next(sizeof(u_long));
	  if(lng == 0) return;
	  printf("- %lu\n", *lng);
	  }
	indent_level--;	
	break;
	}
	
      case TCPOPT_CC:
	printf("[cc]\n");
	print_tcp_ttcp_option();
	break;

      case TCPOPT_CCNEW:
	printf("[cc.new]\n");
	print_tcp_ttcp_option();
	break;

      case TCPOPT_CCECHO:
	printf("[cc.echo]\n");
	print_tcp_ttcp_option();
	break;

      default:
	unsupported++;
	printf("[*unsupported*]\n");
	print_tcp_generic_tcp_option();
      }
    indent_level--;
    } while((position() - pos) < remaining_hdr && *type != TCPOPT_EOL);
  if(*type == TCPOPT_EOL) 
    iprintf(hdr, "option: %hu [end of option list]\n", *type);
  }


/* print a tcp header */
void print_tcp(void)
  {
  int padding = 0;
  int fl = 0;
  const u_long tcp_start = position();
  const struct tcphdr *tcp = 
    (const struct tcphdr *)next(sizeof(struct tcphdr));

  if(tcp == 0) return;
  print_port(hdr, "source port", ntohs(tcp->th_sport), "tcp");
  printf("\n");
  print_port(hdr, "destination port", ntohs(tcp->th_dport), "tcp");
  printf("\n");
  iprintf(hdr, "sequence number: %lu\n", ntohl(tcp->th_seq));
  iprintf(hdr, "acknowledgement number: %lu\n", ntohl(tcp->th_ack));
  iprintf(hdr, "data offset: %hu [%hu octets]\n", tcp->th_off, 
	  tcp->th_off * 4);
  /* print flags */
  iprintf(hdr, "flags: %hu", tcp->th_flags);
  if(tcp->th_flags & TH_FLAGS) printf(" [");
  if(tcp->th_flags & TH_URG) printf("%surg", fl++ ? "|" : "");
  if(tcp->th_flags & TH_ACK) printf("%sack", fl++ ? "|" : "");
  if(tcp->th_flags & TH_PUSH) printf("%spsh", fl++ ? "|" : "");
  if(tcp->th_flags & TH_RST) printf("%srst", fl++ ? "|" : "");
  if(tcp->th_flags & TH_SYN) printf("%ssyn", fl++ ? "|" : "");
  if(tcp->th_flags & TH_FIN) printf("%sfin", fl++ ? "|" : "");
  if(tcp->th_flags & TH_FLAGS) printf("]");
  printf("\n");
  iprintf(hdr, "window: %hu [octet%s]\n", ntohs(tcp->th_win),
	  ntohs(tcp->th_win) == 1 ? "" : "s");
  iprintf(hdr, "checksum: %hu\n", ntohs(tcp->th_sum));
  iprintf(hdr, "urgent pointer: %hu\n", ntohs(tcp->th_urp));

  /* print options, if present */
  if(tcp->th_off * 4 > 20) print_tcp_options(tcp->th_off * 4 - 20);

  /* print padding */
  padding = (position() - tcp_start) % 4;
  if(padding)
    {
    iprintf(hdr, "padding: [%hu octet%s]", padding, padding == 1 ? "" : "s");
    if(verbose)
      {
      /* print padding contents */
      indent_level++;
      print_raw(hdr, padding);
      indent_level--;
      }
    else
      {
      /* skip padding */
      const u_char *skip = next(padding);
      printf(" [omitted]\n");
      skip = 0;
      }
    }

  /* print start of data (until we run out) */
  if(left())
    {
    iprintf(hdr, "data: [%hu octet%s]", left(), left() == 1 ? "" : "s");
    if(verbose)
      {
      printf("\n");
      indent_level++;
      print_raw(hdr, left());
      indent_level--;
      }
    else printf(" [omitted]\n");
    }
  }
